/*
 * Copyright (c) 2001 by George White Consulting
 *	    All Rights Reserved
 * 
 * Anyone who uses this code does so at their own risk. No
 * warranty of any kind is expressed or implied.  Any
 * commercial use of this code is prohibited unless agreed
 * to in writing.
 */
/*
 * Loadable kernel module to add 'hlt' functionality for Intel SMP.
 * Use of this module results in about a 20-25 watt, per CPU, power
 * reduction, when the machine is 100% idle.  At 50% idle, the
 * savings are about half.
 *
 * This module is a replacement for the Solaris idle loop and is
 * invoked by the first call to set_idle_cpu() from the multi-cpu
 * idle loop.
 *
 * Use of this module requires that kernel patch 108520-09 or later
 *  be installed, otherwise, cpu_idle will not point to idle_and_wait(),
 *  so no power savings will be evident.  The routine idle_and_wait()
 *  executes the 'hlt' instruction and seems to have an additional
 *  test of runnable, but I can't be sure since assym.h is missing from
 *  the foundation source release (as is locore.s and a few other critical
 *  pieces).  If patch 108520-09 is not installed, uncomment the define
 *  of USE_MY_IDLE below.
 * An assembly version of idle_and_wait() for Sparc would make this
 *  work for it as well.
 *
 * The Real Time scheduling impact of this is unknown.
 */
#define _KERNEL
#define _SUNDDI

#include <sys/systm.h>
#include <sys/modctl.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/cpuvar.h>
#include <sys/cyclic.h>

/* #define DEBUG		/* print diagnostic messages */
/* #define USE_MY_IDLE		/* for use when 108520-09 is not installed */

typedef void func();

#define HLT_MAXCPU  8
/*
 * This is how often we can poke the other CPU's from the bootcpu.
 *  Might as well poke synchronous to every clock tick since 
 *  bootcpu will be waking up anyway and this will only wake
 *  up other CPU's with something to do.
 */
#define HLT_POKE    NANOSEC / 100	/* .01 second */

struct hlt_data {	/* private data */
    struct modlmisc Modld;
    struct modlinkage Modl;
    int Installed;
    int Numinstalled;
    int Bootcpu;
    int Freqscale;
#ifdef USE_MY_IDLE
    func *Oldidle;
#endif
    func *Oldsetidle;
    cyclic_id_t Hltcyclic;
    int Inidleloop[HLT_MAXCPU];
    int Inhltlim[HLT_MAXCPU];
#ifdef DEBUG
    int Report;
    int Inhltmaxlim[HLT_MAXCPU];
    int Numoldidle[HLT_MAXCPU];
    int Numhltidle[HLT_MAXCPU];
    int Numxtridle[HLT_MAXCPU];
    int Numpoke[HLT_MAXCPU];
#endif
};
#define modld		(hlt_data.Modld)
#define modl		(hlt_data.Modl)
#define installed	(hlt_data.Installed)
#define numinstalled	(hlt_data.Numinstalled)
#define oldidle		(hlt_data.Oldidle)
#define oldsetidle	(hlt_data.Oldsetidle)
#define bootcpu		(hlt_data.Bootcpu)
#define freqscale	(hlt_data.Freqscale)
#define hltcyclic	(hlt_data.Hltcyclic)
#define inidleloop	(hlt_data.Inidleloop)
#define inhltlim	(hlt_data.Inhltlim)
#ifdef DEBUG
#define report		(hlt_data.Report)
#define inhltmaxlim	(hlt_data.Inhltmaxlim)
#define numoldidle	(hlt_data.Numoldidle)
#define numhltidle	(hlt_data.Numhltidle)
#define numxtridle	(hlt_data.Numxtridle)
#define numpoke		(hlt_data.Numpoke)
#endif

extern func *psm_set_idle_cpuf;
static struct hlt_data hlt_data;
static void new_mp_idle(), hc_handler(), hc_poker(), pwr_down_idle();
void (*idle_cpu)();
int cpu_freq;

int _init(void)
{
    int i, t;
    struct cpu *cp, *bootcp;
    cyc_handler_t   hdlr;
    cyc_time_t	when;

    modld.misc_modops = &mod_miscops;
    modld.misc_linkinfo = "SMP 'hlt' support";
    modl.ml_rev = MODREV_1;
    modl.ml_linkage[0] = &modld;
    modl.ml_linkage[1] = NULL;

    hdlr.cyh_func = (cyc_func_t)hc_handler;
    hdlr.cyh_level = CY_HIGH_LEVEL;
    hdlr.cyh_arg = NULL;

    when.cyt_when = (hrtime_t) 0;
    when.cyt_interval = (hrtime_t) HLT_POKE;
    if(ncpus > HLT_MAXCPU) {
	cmn_err(CE_NOTE, "HLTMOD: more than %d cpus is not supported",
	    HLT_MAXCPU);
	t = EINVAL;
    } else if(ncpus == 1) {
#ifndef USE_MY_IDLE
	cmn_err(CE_NOTE, "HLTMOD: requires multiple CPU's be installed");
	t = EINVAL;
#endif
    } else
	t = mod_install(&modl);
    if(!t) {
#ifdef USE_MY_IDLE
	if(ncpus == 1) {    /* just install pwr_down_idle() */
	    mutex_enter(&cpu_lock);	/* not really necessary */
	    oldidle = idle_cpu;		/* this may not be 100% safe */
	    idle_cpu = pwr_down_idle;	/* however, the multi-cpu one is */
	    installed = 2;
	    mutex_exit(&cpu_lock);
	    return(t);
	}
#endif
/*
 * Add a cyclic to the bootcpu which, synchronously with time tics,
 *  pokes the other processors if they are in 'hlt' and have pending work.
 * The main clock cyclic is already attached to the bootcpu and
 *  fires at a 100 per second rate and this is synchronous to it.
 * An integrated kernel solution would call hc_handle() from clock().
 */
	mutex_enter(&cpu_lock);
	bootcpu = getbootcpuid();
	bootcp = cpu_get(bootcpu);
	hdlr.cyh_arg = (void *) bootcpu;
	hltcyclic = cyclic_add(&hdlr, &when);
	cyclic_bind(hltcyclic, bootcp, NULL);
	oldsetidle = psm_set_idle_cpuf;
#ifdef USE_MY_IDLE
	oldidle = idle_cpu;
	idle_cpu = pwr_down_idle;
#endif
	psm_set_idle_cpuf = new_mp_idle;
	for(i = 0; i < ncpus; i++) {
	    inidleloop[i] = 0;
	    inhltlim[i] = 1;
	}
	freqscale = cpu_freq / 30;  /* empirical */
	if(freqscale < 2) freqscale = 2;

	installed = 1;
	mutex_exit(&cpu_lock);
    }
#ifdef DEBUG
    cmn_err(CE_NOTE, "HLTMOD: _init() = %d", t);
#endif
    return(t);
}

int _fini(void)
{
    int i, t;
    struct cpu *cp;

    if(installed == 2) {
#ifdef USE_MY_IDLE
	t = mod_remove(&modl);
	if(!t) {
	    mutex_enter(&cpu_lock); /* not really necessary */
	    idle_cpu = oldidle;
	    oldidle = 0;
	    mutex_exit(&cpu_lock);
	}
#endif
    } else if(installed) {
	psm_set_idle_cpuf = oldsetidle;	/* make sure it won't start back up */
	installed = 0;	    /* causes idle loops to exit and fall back */
	cyclic_remove(hltcyclic);
	i = 1000;
 /* wait for idle loops to fall back, max iter in case numinstalled breaks */
	while((i-- > 0) && (numinstalled > 0))
	    yield();	    /* need to yield to next thread to pop idle loop */

	t = mod_remove(&modl);
	if(!t) {
	    mutex_enter(&cpu_lock); /* probably no longer necessary */
	    oldsetidle = 0;
#ifdef USE_MY_IDLE
	    idle_cpu = oldidle;
	    oldidle = 0;
#endif
	    mutex_exit(&cpu_lock);
	}
    }
#ifdef DEBUG
    cmn_err(CE_NOTE, "HLTMOD: _fini() = %d", t);
#endif
    return(t);
}

int _info(struct modinfo *mp)
{
    return(mod_info(&modl, mp));
}
/*
 * This replaces the multi-processor idle loop of Solaris.  It is
 * implemented by replacing the set_idle_cpu() routine in psm.  To
 * uninstall it, installed is cleared and the new idle loop returns.
 */
static void new_mp_idle(cpid)
int cpid;
{
    struct cpu *cp = CPU;
    int *inip, *nrun;

    if(!installed) return;
    numinstalled++;
    inip = &inidleloop[cpid];
    while(installed) {
#ifdef DEBUG
	if(report <= 0) {
	    int i;
/*	    report = 3000;	/* 30 seconds */
	    report = 60000;	/* 10 minutes */
	    for(i = 0; i < HLT_MAXCPU; i++) {
		if(!numhltidle[i]) continue;
		cmn_err(CE_NOTE, "HLTMOD: cpu=%d, hlt = %d, old = %d, xtr = %d",
			i, numhltidle[i], numoldidle[i], numxtridle[i]);
		cmn_err(CE_CONT, "poke=%d, lim=%d, max=%d",
			numpoke[i], inhltlim[i], inhltmaxlim[i]);
		numhltidle[i] = numoldidle[i] = numxtridle[i] = numpoke[i] = 0;
		inhltmaxlim[i] = 0;
	    }
	}
#endif
/*
 * Once broken out of idle_and_wait, run a flow through idle loop for
 *  inhltlim[cpid] passes through the loop, where inhltlim[cpid] is 
 *  determined adaptively by hc_handler().
 * Will cp->cpu_disp ever change?  If not, move this dereference out
 *  of the inner loop, otherwise, add a mechanism to update the pointer 
 *  when cpu info changes.
 */
	nrun = (int *) &(cp->cpu_disp.disp_nrunnable);
	if(*nrun)
	    swtch();
	if(!*nrun)		/* all swtch() return here */
	    hc_poker(cpid);	/* poke other cpus if halted and runnable */
	if(*inip > 1) {	    /* *inip > 0 we are in flow through idle loop */
	    *inip -= 1;
#ifdef DEBUG
	    numoldidle[cpid]++;
#endif
	    CPU_STAT_ADDQ(cp, cpu_sysinfo.idlethread, 1);
/*
 *	    if(!*nrun && !(cp->cpu_flags & CPU_OFFLINE)) {
 *
 * disp_getwork() is local to the dispatcher, so it is unavailable here.
 * Probably not a real big impact since this is done from disp() as well.
 * Likely to have more impact at higher levels of activity.
 *
 *		    kthread_id_t    t;
 *		    if((t = disp_getwork(cp)) != NULL) {
 *			if(t->t_proc_flag & TP_MSACCT)
 *			    restore_mstate(t);
 *			swtch_to(t);
 *		    }
 *		}
 *	    }
 */
	} else {
	    *inip = 0;	    /* *inip == 0 -> we are in 'hlt' idle loop */
	    if(*nrun == 0) {	/* close enough anyway ... */
		for(;;) {	/* minimize nrun to idle_and_wait() path */
		    (*idle_cpu)();
/*
 * Power down quiesced CPU's.
 */
		    if(!(cp->cpu_flags & CPU_QUIESCED)) break;
		}
		*inip = inhltlim[cpid];
#ifdef DEBUG
		numhltidle[cpid]++;
#endif
/*
 * maybe this deserves its own sysinfo stat
 */
		CPU_STAT_ADDQ(cp, cpu_sysinfo.idlethread, 1);
	    } else {
		*inip = 1;
#ifdef DEBUG
		numxtridle[cpid]++;
#endif
		CPU_STAT_ADDQ(cp, cpu_sysinfo.idlethread, 1);
	    }
	}
    }
    numinstalled--;
}
/*
 * This is implemented as a fault tolerant SMP design.  No mutexes are
 * put around any of the variables, since if they are wrong, all that
 * happens is there might a few extra or fewer flow through idle passes.
 *
 * The cyclic poking of each processor insures that they wake up at a rate
 *  sufficient such that there is no impact on the interactive response
 *  and provides a point to gather statistics.
 *
 * This handler is attached to a cyclic and pokes each CPU (other than
 *  the bootcpu), if its in 'hlt' idle and it has a thread ready to run.
 *  This thread ordinarily runs on the bootcpu, but is should work on
 *  any processor.
 *
 * No mutex is used for cpu[] either, normally, cpu_get() within a
 *  a cpu_lock mutex would be used.  Here, all that really matters is
 *  that cpu[cpid] is a valid address.  It doesn't really matter what 
 *  disp_runnable is, except for performance reasons.  It might be
 *  desirable to lock this out when in cpu_safe mode during p_online
 *  and p_offline, although a test with a loop of online/offline
 *  operations occurring as fast as possible didn't seem to break.
 *
 * Probably better to just call this from clock() instead of using
 *  a seperate cyclic.
 */
static void hc_handler(cpid)
int cpid;
{ 
    int i, j, fs, *inhl;
    struct cpu *cp;

#ifdef DEBUG
    if(report > 0) report -= 1;
#endif
    if(!installed) return;

    fs = freqscale;
    for(i = 0; i < ncpus; i++) {
	cp = cpu[i];
	if(!cp) continue;
	inhl = &inhltlim[i];
	j = cp->cpu_disp.disp_nrunnable;
/*
 * Increase the number of idle loop passes before going back to 'hlt'
 *  idle.  The more threads that are ready to run, the greater the 
 *  increase.  The optimum scale factor was determined empirically
 *  and set in the init code.  As it increases, the power consumption
 *  at low activity will rise as will the low activity response times.
 * Rate of increase needs to be proportional to CPU clock rate.  This
 *  is because interactive interrupts arrive at the same rate, regardless
 *  of the clock frequency and we need to stay out of hlt idle in order
 *  to catch the next interrupt.  If this ramps up too slow, the
 *  mouse gets jumpy and characters are slow to catch up to fast
 *  typing and fast auto repeat rates get jumpy.  This doesn't need
 *  to be as big of a ramp for network or disk related interrupts.
 * The only thing I notice is that when scrolling real fast in emacs,
 *  the display lags behind the cursor a little more than usual.
 */
	if(j) {
	    *inhl += j * fs;
#ifdef DEBUG
	    if(*inhl > inhltmaxlim[i])
		inhltmaxlim[i] = *inhl;
	    numpoke[i] += 1;
#endif
	    if((i != cpid) && !inidleloop[i])
		  poke_cpu(i);
	    continue;
	}
/*
 * Here, we checked the CPU and nothing was ready.  Decrease the number
 *  of idle passes required before entering 'hlt' idle.  Again, the
 *  characteristic of the ramp down is empirical.  The higher ramp down
 *  at larger values of inhl is an adaption to faster CPU's, where inhl
 *  gets real big real fast.
 */
	if(*inhl > 64) *inhl -= 32;
	else if(*inhl > 16) *inhl -= 8;
	else if(*inhl > 4) *inhl -= 2;
	else *inhl = 1;
    }
}
/*
 * Don't count these pokes in the debug stats, they are wakeup pokes
 * and invoked after a processor has completed all of its runnable
 * threads and has nothing else to do.
 */
static void hc_poker(cpid)
int cpid;
{ 
    int i;
    struct cpu *cp;

    for(i = 0; i < ncpus; i++) {
	if(i == cpid) continue;
	cp = cpu[i];
	if(!cp) continue;
	if(!inidleloop[i]) {	/* processor is in 'hlt' wait */
	    if(cp->cpu_disp.disp_nrunnable)
		poke_cpu(i);
	}
    }
}
/*
 * Probably should do a test of runnable with interrupts off
 * and turn interrupts back on immediately before 'hlt'.
 */
static void pwr_down_idle()
{
    asm("hlt");
}
